#!/usr/bin/env python
"""
Usage:
  mmi-runner <engine> <configfile> [-o <outputvar>...] [-g <globalvar>...] [--interval <interval>] [--disable-logger] [--pause] [--mpi <method>] [--track <server>] [--port <port>] [--bmi-class <bmi-class>]
  mmi-runner -h | --help

Positional arguments:
  engine model  engine
  configfile    model configuration file

Optional arguments:
  -h, --help               show this help message and exit
  --interval <interval>    publish results every <interval> timesteps [default: 1]
  -o <outputvar>           output variables, will be broadcasted each <interval> timestep
  --disable-logger         do not inject logger into the BMI library
  --pause                  start in paused mode, send update messages to progress [default: False]
  --mpi <method>           communicate with mpi nodes using one of the methods: root (communicate with rank 0), all (one socket per rank)
  --port <port>            "random" or integer base port, port is computed as req/rep = port + rank*3 + 0, push/pull = port + rank*3 + 1, pub/sub = port + rank*3 + 2 [default: random]
  --track <tracker>        server to subscribe to for tracking
  --bmi-class <bmi-class>  when used the engine is assumed to be the full name of a Python class that implements bmi [default: bmi.wrapper.BMIWrapper]
"""  # noqa: E501


import warnings
import logging

import docopt

import mmi.runner

logger = logging.getLogger(__name__)


def main():
    """run mmi runner"""
    logging.basicConfig()
    logger.info("mmi-runner")
    warnings.warn(
        "You are using the mmi-runner script, please switch to `mmi runner`",
        DeprecationWarning
    )
    arguments = docopt.docopt(__doc__)
    kwargs = parse_args(arguments)
    runner = mmi.runner.Runner(
        **kwargs
    )
    runner.run()


def parse_args(arguments, wrapper_kwargs={}):
    """
    MMI Runner
    """
    # make a socket that replies to message with the grid

    # if we are running mpi we want to know the rank
    args = {}
    positional = [
        'engine',
        'configfile',
    ]
    for key in positional:
        args[key] = arguments['<' + key + '>']

    #  integer if not 'random'
    port = arguments['--port']
    args['port'] = port if port == 'random' else int(port)

    # integer, default 1
    interval = arguments['--interval']
    args['interval'] = int(interval) if interval else 1
    # boolean
    args['pause'] = bool(arguments['--pause'])
    args['bmi_class'] = arguments['--bmi-class']
    args['output_vars'] = arguments['-o']
    args['tracker'] = arguments['--track']

    return args


if __name__ == '__main__':
    main()



import requests
from six.moves.urllib.parse import urljoin


class MMITracker(object):
    def __init__(self, tracker_url):
        """
        The 'database' has module uuids as key and its metadata as value.
        """
        self.tracker_url = tracker_url
        self.database = {}

    def update(self):
        """Update myself with the tracker server"""
        self.database = requests.get(self.tracker_url).json()

    def key_occurrence(self, key, update=True):
        """
        Return a dict containing the value of the provided key and its uuid as
        value.
        """
        if update:
            self.update()
        result = {}
        for k, v in self.database.items():
            if key in v:
                result[str(v[key])] = k
        return result

    def zmq_address(self, key):
        """
        Return a ZeroMQ address to the module with the provided key.
        """
        zmq_address = "tcp://" + self.database[key]['node'] + ":" + str(self.database[key]['ports']['REQ'])
        return zmq_address

    def unregister(self, uuid):
        """Unregister a uuid from tracker"""
        requests.delete(
            urljoin(self.tracker_url, 'models' + "/" + uuid)
        )



import urllib.parse

import zmq

from bmi.api import IBmi

from mmi import send_array, recv_array


class MMIClient(IBmi):
    """Client to talk to MMI server, using the same API as BMI"""
    def __init__(self, zmq_address, poll_timeout=10000, zmq_flags=0):
        """
        Constructor
        """
        # Open ZeroMQ socket
        context = zmq.Context()

        self.sockets = {}

        host, port = urllib.parse.splitport(zmq_address)

        ports = {
            zmq.REQ: int(port),
            zmq.PUSH: int(port) + 1,
            zmq.SUB: int(port) + 2
        }

        for socket, port in ports.items():
            self.sockets[socket] = context.socket(socket)
            self.sockets[socket].connect(
                ":".join([
                    host,
                    str(port)
                ])
            )

        self.pollers = {}
        for socket in [zmq.REQ, zmq.PUSH, zmq.SUB]:
            poller = zmq.Poller()
            self.pollers[socket] = poller
            self.pollers[socket].register(self.sockets[socket], zmq.POLLIN)

        self.poll_timeout = poll_timeout
        self.zmq_flags = zmq_flags

        self.socket = self.sockets[zmq.REQ]
        self.poll = self.pollers[zmq.REQ]

    def _close_sockets(self):
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.close()
        self.poll.unregister(self.socket)

    # from here: BMI commands that get translated to MMI.
    def initialize(self, configfile=None):
        """
        Initialize the module
        """

        method = "initialize"

        A = None
        metadata = {method: configfile}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

    def finalize(self):
        """
        Finalize the module
        """

        method = "finalize"

        A = None
        metadata = {method: -1}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

    def get_var_count(self):
        """
        Return number of variables
        """

        method = "get_var_count"

        A = None
        metadata = {method: -1}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_var_name(self, i):
        """
        Return variable name
        """

        method = "get_var_name"

        A = None
        metadata = {method: i}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_var_type(self, name):
        """
        Return variable name
        """

        method = "get_var_type"

        A = None
        metadata = {method: name}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_var_rank(self, name):
        """
        Return variable rank
        """

        method = "get_var_rank"

        A = None
        metadata = {method: name}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_var_shape(self, name):
        """
        Return variable shape
        """

        method = "get_var_shape"

        A = None
        metadata = {method: name}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_var(self, name):
        """
        Return an nd array from model library
        """

        method = "get_var"

        A = None
        metadata = {method: name}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return A

    def set_current_time(self, t):
        """
        Set current time of simulation
        """

        method = "set_current_time"

        A = None
        metadata = {method: t}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

    def set_var(self, name, var):
        """
        Set the variable name with the values of var
        """

        method = "set_var"

        A = var
        metadata = {method: name}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

    def get_start_time(self):
        """
        Return start time
        """

        method = "get_start_time"

        A = None
        metadata = {method: -1}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_end_time(self):
        """
        Return end time of simulation
        """

        method = "get_end_time"

        A = None
        metadata = {method: -1}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def get_current_time(self):
        """
        Return current time of simulation
        """

        method = "get_current_time"

        A = None
        metadata = {method: -1}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def update(self, dt):
        """
        Advance the module with timestep dt
        """

        method = "update"

        A = None
        metadata = {method: dt}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

    # TODO:  Do we really need these two?
    def inq_compound(self, name):
        """
        Return the number of fields of a compound type.
        """
        pass

    def inq_compound_field(self, name, index):
        """
        Lookup the type,rank and shape of a compound field
        """
        pass

    def remote(self, action):
        """
        Function specific for MMI, not BMI.
        action is one of: "play", "stop", "pause", "rewind", "quit"
        """
        method = "remote"

        A = None
        metadata = {method: action}

        send_array(self.socket, A, metadata)
        A, metadata = recv_array(
            self.socket, poll=self.poll, poll_timeout=self.poll_timeout,
            flags=self.zmq_flags)

        return metadata[method]

    def subscribe(self, topic=b''):
        """subscribe to the SUB socket, to listen for incomming variables"""
        self.sockets[zmq.SUB].setsockopt(zmq.SUBSCRIBE, topic)



# -*- coding: utf-8 -*-

"""Console script for mmi."""

import logging
import sys

import click
import tornado.ioloop

import mmi.runner
import mmi.tracker

logging.basicConfig()
logger = logging.getLogger(__name__)


@click.group()
def cli(args=None):
    """Console script for mmi."""
    click.echo("Model Message Interface")
    return 0


@cli.command()
def tracker():
    """start a tracker to register running models"""
    application = mmi.tracker.app()
    application.listen(22222)
    logger.info('serving at port 22222')
    tornado.ioloop.IOLoop.instance().start()


@cli.command()
@click.argument('engine')
@click.argument(
    'configfile',
    type=click.Path(exists=True)
)
@click.option(
    '--output',
    '-o',
    'output_vars',
    multiple=True,
    help='output variables, will be broadcasted each <interval> timestep'
)
@click.option(
    '--interval',
    type=int,
    default=1,
    help='publish results every <interval> timesteps'
)
@click.option(
    '--pause',
    is_flag=True,
    help='start in paused mode, send update messages to progress'
)
@click.option(
    '--mpi',
    type=click.Choice(['root', 'all']),
    help="communicate with mpi nodes using one of the methods: root (communicate with rank 0), all (one socket per rank)"
)
@click.option('--tracker', help='server to subscribe to for tracking')
@click.option(
    '--port',
    type=int,
    help='''"random" or integer base port,
    port is computed as req/rep = port + rank*3 + 0, push/pull = port + rank*3 + 1, pub/sub = port + rank*3 + 2 [default: random]
    '''
)
@click.option(
    '--bmi-class',
    'bmi_class',
    default='bmi.wrapper.BMIWrapper',
    help='the full name of a python class that implements bmi'
)
def runner(
        engine,
        configfile,
        output_vars,
        interval,
        pause,
        mpi,
        tracker,
        port,
        bmi_class
):
    """
    run a BMI compatible model
    """
    # keep track of info
    # update mpi information or use rank 0
    runner = mmi.runner.Runner(
        engine=engine,
        configfile=configfile,
        output_vars=output_vars,
        interval=interval,
        pause=pause,
        mpi=mpi,
        tracker=tracker,
        port=port,
        bmi_class=bmi_class
    )
    runner.run()


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import importlib
import os
import logging
import json
from six.moves.urllib.parse import urljoin
import itertools
import atexit
import platform

import requests
import zmq
import zmq.eventloop.zmqstream
from zmq.eventloop import ioloop


import bmi.wrapper

from mmi import send_array, recv_array

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

ioloop.install()


class Runner(object):
    def __init__(
            self,
            engine,
            configfile,
            output_vars=None,
            interval=1,
            tracker=None,
            mpi=False,
            port='random',
            bmi_class='bmi.wrapper.BMIWrapper',
            *args,
            **kwargs
    ):
        # store options
        self.engine = engine
        self.configfile = configfile
        if output_vars is None:
            self.output_vars = []
        else:
            self.output_vars = output_vars
        self.interval = interval
        self.tracker = tracker
        self.mpi = mpi
        self.port = port
        self.bmi_class = bmi_class
        self.args = args
        self.kwargs = kwargs

        # load mpi
        self.mpi_info = self.initialize_mpi(mpi)
        self.ports = self.create_ports(port, mpi=mpi, rank=self.mpi_info['rank'])
        self.sockets = self.create_sockets()

        self.model = self.create_bmi_model(engine, bmi_class)
        self.metadata = {}
        self.fill_metadata()

    @staticmethod
    def initialize_mpi(mpi=False):
        """initialize mpi settings"""
        if mpi:
            import mpi4py.MPI
            comm = mpi4py.MPI.COMM_WORLD
            rank = comm.Get_rank()
            size = comm.Get_size()
        else:
            comm = None
            rank = 0
            size = 1
        return {
            "comm": comm,
            "rank": rank,
            "size": size,
            "mode": mpi
        }

    @staticmethod
    def create_ports(port, mpi, rank):
        """create a list of ports for the current rank"""
        if port == "random" or port is None:
            # ports will be filled in using random binding
            ports = {}
        else:
            port = int(port)
            ports = {
                "REQ": port + 0,
                "PUSH": port + 1,
                "SUB": port + 2
            }
        # if we want to communicate with separate domains
        # we have to setup a socket for each of them
        if mpi == 'all':
            # use a socket for each rank rank
            for port in ports:
                ports[port] += (rank * 3)
        return ports

    @staticmethod
    def import_from_string(full_class_name):
        """return a class based on it's full class name"""
        s = full_class_name.split('.')
        class_name = s[-1]
        module_name = full_class_name[:-len(class_name)-1]
        module = importlib.import_module(module_name)
        # the class, it's common to spell with k as class is reserved
        klass = getattr(module, class_name)
        return klass

    def create_bmi_model(self, engine, bmi_class=None, wrapper_kwargs=None):
        """initialize a bmi mode using an optional class"""
        if wrapper_kwargs is None:
            wrapper_kwargs = {}
        if bmi_class is None:
            wrapper_class = bmi.wrapper.BMIWrapper
        else:
            wrapper_class = self.import_from_string(bmi_class)
        try:
            """most models use engine as a first argument"""
            model = wrapper_class(
                engine,
                **wrapper_kwargs
            )
        except TypeError as e:
            """but old python engines are engines, so they don't, but they should """
            logger.warn(
                'Model wrapper %s does not accept engine as a first argument',
                wrapper_class
            )
            model = wrapper_class(
                **wrapper_kwargs
            )
        # by default start running
        model.state = 'play'
        return model

    def register(self):
        """register model at tracking server"""
        # connect to tracker
        result = requests.post(urljoin(self.tracker, 'models'), data=json.dumps(self.metadata))
        logger.debug("registered at server %s: %s", self.tracker, result)
        self.metadata["tracker"] = result.json()

    def unregister(self):
        """unregister model at tracking server"""
        uuid = self.metadata["tracker"]["uuid"]
        # connect to server
        result = requests.delete(urljoin(self.tracker, 'models' + "/" + uuid))
        logger.debug("unregistered at server %s with %s: %s", self.tracker, uuid, result)

    def fill_metadata(self):
        self.metadata["node"] = platform.node()
        self.metadata.update({
            "engine": self.engine,
            "configfile": self.configfile,
            "ports": self.ports,
            "mpi": {
                "rank": self.mpi_info['rank'],
                "size": self.mpi_info['size']
            }
        })
        metadata_filename = os.path.join(
            os.path.dirname(self.configfile),
            "metadata.json"
        )
        if os.path.isfile(metadata_filename):
            self.metadata.update(json.load(open(metadata_filename)))
        # update connection information from external service
        # You might want to disable this if you have some sort of sense of privacy
        try:
            self.metadata["ifconfig"] = requests.get("http://ipinfo.io/json").json()
        except IOError:
            pass

    def process_incoming(self):
        """
        process incoming messages

        data is a dict with several arrays
        """

        model = self.model
        sockets = self.sockets

        # Check for new messages
        if not sockets:
            return
        # unpack sockets
        poller = sockets['poller']
        rep = sockets['rep']
        # pull = sockets['pull']
        pub = sockets['pub']
        items = poller.poll(10)
        for sock, n in items:
            for i in range(n):

                A, metadata = recv_array(sock)
                var = None

                # bmi actions
                if "update" in metadata:
                    dt = float(metadata["update"])
                    model.update(dt)
                    metadata["dt"] = dt
                elif "get_var" in metadata:
                    name = metadata["get_var"]
                    # temporary implementation
                    if metadata.get("copy", False):
                        var = model.get_var(name).copy()
                    else:
                        var = model.get_var(name)
                    if var is None:
                        logger.warning("Get_var returns None for %s" % name)
                    metadata['name'] = name
                    # assert socket is req socket
                elif "get_var_count" in metadata:
                    # temporary implementation
                    n = model.get_var_count()
                    metadata['get_var_count'] = n
                    # assert socket is req socket
                elif "get_var_rank" in metadata:
                    # temporary implementation
                    var_name = metadata['get_var_rank']
                    n = model.get_var_rank(var_name)
                    metadata['get_var_rank'] = n
                    # assert socket is req socket
                elif "get_var_shape" in metadata:
                    # temporary implementation
                    var_name = metadata['get_var_shape']
                    n = model.get_var_shape(var_name)
                    metadata['get_var_shape'] = tuple([int(item) for item in n])
                    # assert socket is req socket
                elif "get_var_type" in metadata:
                    # temporary implementation
                    var_name = metadata['get_var_type']
                    n = model.get_var_type(var_name)
                    metadata['get_var_type'] = n
                    # assert socket is req socket
                elif "get_var_name" in metadata:
                    i = int(metadata["get_var_name"])
                    name = model.get_var_name(i)
                    metadata['get_var_name'] = name
                    # assert socket is req socket
                elif "set_current_time" in metadata:
                    t = float(metadata["set_current_time"])
                    model.set_current_time(t)
                elif "set_var" in metadata:
                    name = metadata["set_var"]
                    # logger.debug("setting variable %s", name)
                    model.set_var(name, A)
                    metadata["name"] = name  # !?
                elif "set_var_slice" in metadata:
                    name = metadata["set_var_slice"]
                    logger.debug("setting variable %s", name)
                    start = metadata["start"]
                    count = metadata["count"]
                    model.set_var_slice(name, start, count, A)
                    metadata["name"] = name  # !?
                elif "set_var_index" in metadata:
                    name = metadata["set_var_index"]
                    logger.debug("setting variable %s using index index", name)
                    index = metadata["index"]
                    # TODO: test if this is fast enough.
                    # Otherwise move to BMI ++ but that is
                    # a bit of a burden on implementers
                    var = model.get_var(name).copy()
                    var.flat[index] = A
                    model.set_var(name, var)
                    metadata["name"] = name  # !?
                elif "get_current_time" in metadata:
                    metadata["get_current_time"]
                    t = model.get_current_time()
                    metadata['get_current_time'] = t
                elif "get_time_step" in metadata:
                    metadata["get_time_step"]
                    dt = model.get_time_step()
                    metadata['get_time_step'] = dt
                elif "get_end_time" in metadata:
                    metadata["get_end_time"]
                    t = model.get_end_time()
                    metadata['get_end_time'] = t
                elif "get_start_time" in metadata:
                    metadata["get_start_time"]
                    t = model.get_start_time()
                    metadata['get_start_time'] = t
                    # assert socket is req socket
                # custom actions
                elif "remote" in metadata:
                    assert metadata["remote"] in {
                        "play", "stop", "pause", "rewind", "quit"}
                    model.state = metadata["remote"]
                elif "operator" in metadata:
                    # TODO: support same operators as MPI_ops here....,
                    # TODO: reduce before apply
                    # TODO: assert pull socket
                    pass
                    # S = tuple(slice(*x) for x in action['slice'])
                    # print(repr(arr[S]))
                    # if action['operator'] == 'setitem':
                    #     arr[S] = data
                    # elif action['operator'] == 'add':
                    #     arr[S] += data
                elif "initialize" in metadata:
                    config_file = metadata["initialize"]
                    model.initialize(str(config_file))
                elif "finalize" in metadata:
                    model.finalize()
                else:
                    logger.warn("got unknown message {} from socket {}".format(str(metadata), sock))
                if sock.socket_type == zmq.REP:
                    # reply
                    send_array(rep, var, metadata=metadata)
                # any getter requested through the pull socket?
                elif any(x.startswith("get_") for x in metadata) and sock.socket_type == zmq.PULL:
                    # return through the pub socket
                    send_array(pub, var, metadata=metadata)

    def create_sockets(self):
        """create zmq sockets"""

        ports = self.ports

        context = zmq.Context()
        poller = zmq.Poller()

        # Socket to handle init data
        rep = context.socket(zmq.REP)
        # this was inconsequent: here REQ is for the client, we reply with REP.
        # PULL and PUB is seen from here, not from the client.
        # Is now renamed to PUSH and SUB: everything is seen from outside.
        if "REQ" in ports:
            rep.bind(
                "tcp://*:{port}".format(port=ports["REQ"])
            )
        else:
            ports["REQ"] = rep.bind_to_random_port(
                "tcp://*"
            )

        pull = context.socket(zmq.PULL)
        if "PUSH" in ports:
            pull.bind(
                "tcp://*:{port}".format(port=ports["PUSH"])
            )
        else:
            ports["PUSH"] = pull.bind_to_random_port(
                "tcp://*"
            )

        # for sending model messages
        pub = context.socket(zmq.PUB)
        if "SUB" in ports:
            pub.bind(
                "tcp://*:{port}".format(port=ports["SUB"])
            )
        else:
            ports["SUB"] = pub.bind_to_random_port(
                "tcp://*"
            )

        poller.register(rep, zmq.POLLIN)
        poller.register(pull, zmq.POLLIN)
        sockets = dict(
            poller=poller,
            rep=rep,
            pull=pull,
            pub=pub
        )
        return sockets

    def run(self):
        """run the model"""

        model = self.model
        configfile = self.configfile
        interval = self.interval
        sockets = self.sockets

        model.initialize(configfile)
        if model.state == 'pause':
            logger.info(
                "model initialized and started in pause mode, waiting for requests"
            )
        else:
            logger.info("model started and initialized, running")

        if self.tracker:
            self.register()
            atexit.register(self.unregister)

        self.process_incoming()

        # Keep on counting indefinitely
        counter = itertools.count()
        logger.info("Entering timeloop...")
        for i in counter:
            while model.state == "pause":
                # keep waiting for messages when paused
                # process_incoming should set model.state to play
                self.process_incoming()
            else:
                # otherwise process messages once and continue
                self.process_incoming()
            if model.state == "quit":
                break

            # lookup dt or use -1 (default)
            dt = model.get_time_step() or -1
            model.update(dt)

            # check counter, if not a multiple of interval, skip this step
            if i % interval:
                continue

            for key in self.output_vars:
                value = model.get_var(key)
                metadata = {'name': key, 'iteration': i}
                # 4ms for 1M doubles
                logger.debug("sending {}".format(metadata))
                if 'pub' in sockets:
                    send_array(sockets['pub'], value, metadata=metadata)

        logger.info("Finalizing...")
        model.finalize()



import logging

import numpy as np
import zmq
import shapely.geometry
import osgeo.osr

from . import send_array, recv_array

logging.basicConfig()
logger = logging.getLogger(__name__)


class Views(object):
    # TODO: rewrite using config file per engine
    @staticmethod
    def grid(context):
        meta = context["value"]
        # Get connection info
        node = meta['node']
        node = 'localhost'
        req_port = meta["ports"]["REQ"]
        ctx = context["ctx"]
        req = ctx.socket(zmq.REQ)
        req.connect("tcp://%s:%s" % (node, req_port))
        # Get grid variables
        send_array(req, metadata={"get_var": "xk"})
        xk, A = recv_array(req)
        send_array(req, metadata={"get_var": "yk"})
        yk, A = recv_array(req)
        # Spatial transform
        points = np.c_[xk, yk]
        logger.info("points shape: %s, values: %s", points.shape, points)
        src_srs = osgeo.osr.SpatialReference()
        src_srs.ImportFromEPSG(meta["epsg"])
        dst_srs = osgeo.osr.SpatialReference()
        dst_srs.ImportFromEPSG(4326)
        transform = osgeo.osr.CoordinateTransformation(src_srs, dst_srs)
        wkt_points = transform.TransformPoints(points)
        geom = shapely.geometry.MultiPoint(wkt_points)

        geojson = shapely.geometry.mapping(geom)
        return geojson


views = Views()
